Attribute VB_Name = "modMem"
Option Explicit
Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pDst As Any, pSrc As Any, ByVal ByteLen As Long)
Declare Sub ZeroMemory Lib "kernel32" Alias "RtlZeroMemory" (pDst As Any, ByVal ByteLen As Long)
Declare Function GetProcessHeap Lib "kernel32" () As Long
Declare Function HeapAlloc Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, ByVal dwBytes As Long) As Long
Declare Function HeapFree Lib "kernel32" (ByVal hHeap As Long, ByVal dwFlags As Long, lpMem As Any) As Long
Dim Heap As Long
Dim GlobMem() As Byte
Dim Stack() As Variant
Dim MaxStack As Long
Const HEAP_ZERO_MEMORY As Long = &H8

Private Function GetAddress(ByVal Address As Long) As Long
  If Address < &H100000 Then Address = VarPtr(GlobMem(Address))
  GetAddress = Address
End Function

Function AllocBytes(ByVal Bytes As Long) As Long
  If Bytes Then AllocBytes = HeapAlloc(Heap, HEAP_ZERO_MEMORY, Bytes)
End Function

Sub UnallocBytes(ByVal Address As Long)
  If Address Then HeapFree Heap, 0, ByVal Address
End Sub

Sub PutMem1(ByVal Address As Long, ByVal Value As Boolean)
  CopyMemory ByVal GetAddress(Address), Value, 1
End Sub

Sub PutMem4(ByVal Address As Long, ByVal Value As Long)
  CopyMemory ByVal GetAddress(Address), Value, Len(Value)
End Sub

Sub PutMem8(ByVal Address As Long, ByVal Value As Double)
  CopyMemory ByVal GetAddress(Address), Value, Len(Value)
End Sub

Sub PutMemStr(ByVal Address As Long, Value As String)
  Dim RealAddr As Long, Length As Long, OldStr As Long

  Length = Len(Value) * 2
  If Length Then
    RealAddr = AllocBytes(Length + 6)
    CopyMemory ByVal RealAddr, Length, 4
    CopyMemory ByVal RealAddr + 4, ByVal StrPtr(Value), Length
  End If
  Address = GetAddress(Address)
  CopyMemory OldStr, ByVal Address, Len(Address)
  If OldStr Then UnallocBytes OldStr
  CopyMemory ByVal Address, RealAddr, Len(RealAddr)
End Sub

Sub PutMemVar(ByVal Address As Long, ByVal Value)
  Dim Temp

  Address = GetAddress(Address)
  CopyMemory Temp, ByVal Address, 16
  CopyMemory ByVal Address, Value, 16
  ZeroMemory Value, 16
End Sub

Function GetMem1(ByVal Address As Long) As Boolean
  Dim TmpByte As Byte

  CopyMemory TmpByte, ByVal GetAddress(Address), Len(TmpByte)
  If TmpByte Then GetMem1 = True
End Function

Function GetMem4(ByVal Address As Long) As Long
  CopyMemory Address, ByVal GetAddress(Address), Len(Address)
  GetMem4 = Address
End Function

Function GetMem8(ByVal Address As Long) As Double
  Dim TmpDbl As Double

  CopyMemory TmpDbl, ByVal GetAddress(Address), Len(TmpDbl)
  GetMem8 = TmpDbl
End Function

Function GetMemStr(ByVal Address As Long) As String
  Dim Size As Long, RealAddr As Long
  Dim TmpStr As String

  CopyMemory RealAddr, ByVal GetAddress(Address), Len(RealAddr)
  If RealAddr = 0 Then Exit Function
  CopyMemory Size, ByVal RealAddr, 4
  If Size = 0 Then Exit Function
  TmpStr = Space$(Size \ 2)
  CopyMemory ByVal StrPtr(TmpStr), ByVal RealAddr + 4, Size
  GetMemStr = TmpStr
End Function

Function GetMemVar(ByVal Address As Long) As Variant
  Dim VarToRet, Temp

  CopyMemory Temp, ByVal GetAddress(Address), 16
  VarToRet = Temp
  ZeroMemory Temp, 16
  GetMemVar = VarToRet
End Function

Sub SetGlobMemSpace(ByVal Bytes As Long)
  Heap = GetProcessHeap()
  If Bytes Then ReDim GlobMem(Bytes - 1)
End Sub

Sub Push(Value As Variant)
  ReDim Preserve Stack(MaxStack)
  Stack(MaxStack) = Value
  MaxStack = MaxStack + 1
End Sub

Function Pop(Optional ByVal DoNotPop As Boolean) As Variant
  If MaxStack = 0 Then Exit Function
  Pop = Stack(MaxStack - 1)
  If DoNotPop Then Exit Function
  MaxStack = MaxStack - 1
  If MaxStack Then ReDim Preserve Stack(MaxStack - 1) Else Erase Stack
End Function

Function UpperBound(ByVal Address As Long) As Long
  If Address Then UpperBound = GetMem4(Address - 4)
End Function

Sub DoRedim(ByVal Address As Long, ByVal ElementSize As Long, ByVal NewSize As Long)
  Dim RealAddress As Long

  RealAddress = GetMem4(Address)
  If RealAddress Then UnallocBytes RealAddress - 4
  If NewSize = 0 Then
    PutMem4 Address, 0
    Exit Sub
  End If
  RealAddress = AllocBytes(ElementSize * NewSize + 4)
  PutMem4 RealAddress, NewSize
  PutMem4 Address, RealAddress + 4
End Sub

Sub DoRedimPreserve(ByVal Address As Long, ByVal ElementSize As Long, ByVal NewSize As Long)
  Dim RealAddress As Long, Addr2 As Long, Ubnd As Long

  Addr2 = GetMem4(Address)
  Ubnd = UpperBound(Addr2)
  If Ubnd = NewSize Then Exit Sub
  If NewSize = 0 Then
    If Addr2 Then UnallocBytes Addr2 - 4
    PutMem4 Address, 0
    Exit Sub
  End If
  RealAddress = AllocBytes(ElementSize * NewSize + 4)
  PutMem4 RealAddress, NewSize
  RealAddress = RealAddress + 4
  If Addr2 Then
    If Ubnd > NewSize Then
      CopyMemory ByVal RealAddress, ByVal Addr2, ElementSize * NewSize
    Else
      CopyMemory ByVal RealAddress, ByVal Addr2, ElementSize * Ubnd
    End If
  End If
  PutMem4 Address, RealAddress
End Sub
